package gov.va.capri;

import org.apache.log4j.Logger;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;


/**
 * An InputStream decorator that reads up to, but not including, the last N
 * number of bytes available on the specified stream. The stream will block the
 * current thread for a configurable number of seconds until the specified
 * authorization request comes back; if authorized, we continue and read the
 * last N bytes. If unsuccessful, an IOException is thrown.
 * 
 * @author vhaisdbarryc
 * 
 */
public class AuthorizedInputStream extends InputStream {

    private static final Logger log = Logger.getLogger(AuthorizedInputStream.class);

    static final int INITIAL_BUFFER_SIZE = 32 * 1024;

	InputStream is;
	Future<Boolean> authorizationFuture;
	int authTimeoutSeconds;
	int trailingBytesToHold;

	BufferedInputStream bis;
	byte[] byteBuffer;

	boolean authorizationAttempted = false;
	boolean authorized = false;

	private List<AuthCallback> authCallbacks = new ArrayList<AuthCallback>();

	public AuthorizedInputStream(InputStream is,
			Future<Boolean> authorizationFuture, int authTimeoutSeconds,
			int trailingBytesToHold) {
		this.is = is;
		this.authorizationFuture = authorizationFuture;
		this.authTimeoutSeconds = authTimeoutSeconds;
		this.trailingBytesToHold = trailingBytesToHold;

		bis = new BufferedInputStream(is, INITIAL_BUFFER_SIZE
				+ trailingBytesToHold);
		byteBuffer = new byte[INITIAL_BUFFER_SIZE + trailingBytesToHold];
	}

	public AuthorizedInputStream addAuthCallback(AuthCallback a) {
		authCallbacks.add(a);
		return this;
	}

	public static interface AuthCallback {
		void authComplete(boolean isAuthorized);
	}

	private static interface ReadImplementation {
		int go(BufferedInputStream bis, byte[] b, int off, int len)
				throws IOException;
	}

	/*
	 * Pre-instantiating implementations for speed (as opposed to having one
	 * with b/off/len fields)
	 */
	ReadImplementation simpleRead = new ReadImplementation() {
		@Override
		public int go(BufferedInputStream bis, byte[] b, int off, int len)
				throws IOException {
			return bis.read();
		}
	};

	ReadImplementation readArray = new ReadImplementation() {
		@Override
		public int go(BufferedInputStream bis, byte[] b, int off, int len)
				throws IOException {
			return bis.read(b);
		}
	};

	ReadImplementation readArrayComplex = new ReadImplementation() {
		@Override
		public int go(BufferedInputStream bis, byte[] b, int off, int len)
				throws IOException {
			return bis.read(b, off, len);
		}
	};

	private int commonRead(int desiredLength, ReadImplementation callback,
			byte[] b, int off, int len) throws IOException {

        log.debug("Desired length = {} bytes... " + (Integer.toString(desiredLength)));

        if (authorizationAttempted)
            if (authorized) {
                int numBytes = callback.go(bis, b, off, len);
                log.debug("Previously authorized, returning {} bytes..." + (Integer.toString(numBytes)));
                return numBytes;
            } else {
                throw new IOException("Authorization previously failed");
            }

        int availableForMark = getActualBytesAvailable(desiredLength);

		/* If no bytes are immediately available, return */
        if (availableForMark == -1) {
            log.debug("no bytes available, returning.");
            return -1;
        }

        if (availableForMark == desiredLength + trailingBytesToHold) {
            int numBytes = callback.go(bis, b, off, len);
            log.debug("Returning the next {} bytes." + Integer.toString(numBytes));
            return numBytes;
        } else {
            try {
                performAuth();
            } catch (IOException e) {
                for (AuthCallback ac : authCallbacks)
                    ac.authComplete(false);
                throw e;
            }

            for (AuthCallback ac : authCallbacks)
                ac.authComplete(authorized);

            if (authorized) {
                log.debug("Authorized.");
                int remaining = callback.go(bis, b, off, len);
                log.debug("Returning the next {} bytes after authorization. " + Integer.toString(remaining));
                return remaining;
            } else {
                log.debug("Unauthorized.");
                throw new AuthorizationFailedException("AuthFailed");
            }
        }
    }

	public int getActualBytesAvailable(int desiredLength) throws IOException {
		log.debug("reserving {} bytes for mark " + Integer.toString(desiredLength + trailingBytesToHold));
		bis.mark(desiredLength + trailingBytesToHold);
		if (byteBuffer.length < desiredLength + trailingBytesToHold)
			byteBuffer = new byte[desiredLength + trailingBytesToHold];

		int numRead = 0, result = 0;
		while (numRead < (desiredLength + trailingBytesToHold) && result != -1) {
			result = bis.read(byteBuffer, 0,
					(desiredLength + trailingBytesToHold) - numRead);
			if (result != -1) {
				numRead += result;
			}
			log.debug("\tRead {} total bytes within mark range " +Integer.toString(numRead));
		}
		log.debug("\tAt least {} still available on stream " +Integer.toString(numRead));
		bis.reset();
		return numRead;
	}

	private void performAuth() throws IOException {
		try {
			authorizationAttempted = true;
			log.debug("Waiting for authorization...");
			authorized = authorizationFuture.get(authTimeoutSeconds, TimeUnit.SECONDS);
		} catch (InterruptedException e) {
			throw new IOException(e);
		} catch (ExecutionException e) {
			log.error("Execution exception while authorizing", e);
			/*
			 * If our actual authorization process threw an exception, propagate
			 * it here since it was unexpected.
			 */
			throw new AuthorizationFailedException("There was an exception while attempting to authorize", e);
		} catch (TimeoutException e) {
			/*
			 * If our process took longer than the configured timeout, we assume
			 * it's broken and can't authorize the user.
			 */
			throw new AuthorizationFailedException("Authorization timed out", e);
		}
	}

	@Override
	public int read() throws IOException {
		return commonRead(1, simpleRead, null, -1, -1);
	}

	@Override
	public int read(byte[] b) throws IOException {
		return commonRead(b.length, readArray, b, -1, -1);
	}

	@Override
	public int read(byte[] b, int off, int len) throws IOException {
		return commonRead(len, readArrayComplex, b, off, len);
	}

	@Override
	public long skip(long n) throws IOException {
		return bis.skip(n);
	}

	@Override
	public int available() throws IOException {
		return bis.available();
	}

	@Override
	public void close() throws IOException {
		bis.close();
	}

	@Override
	public synchronized void mark(int readlimit) {
		bis.mark(readlimit);
	}

	@Override
	public synchronized void reset() throws IOException {
		bis.reset();
	}

	@Override
	public boolean markSupported() {
		return bis.markSupported();
	}

}
