package gov.va.capri;

/**
 * Copyright MITRE
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import gov.va.capri.AuthorizedInputStream.AuthCallback;
import org.apache.http.*;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.methods.AbstractExecutionAwareRequest;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.utils.URIUtils;
import org.apache.http.config.SocketConfig;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLContexts;
import org.apache.http.entity.InputStreamEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClientBuilder;
import org.apache.http.impl.conn.BasicHttpClientConnectionManager;
import org.apache.http.message.BasicHeader;
import org.apache.http.message.BasicHttpEntityEnclosingRequest;
import org.apache.http.message.BasicHttpRequest;
import org.apache.http.message.HeaderGroup;
import org.apache.http.util.EntityUtils;
import org.apache.log4j.Logger;

import javax.net.ssl.SSLContext;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import java.io.*;
import java.net.URI;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.util.BitSet;
import java.util.Enumeration;
import java.util.Formatter;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeoutException;

/**
 * An HTTP reverse proxy/gateway servlet. It is designed to be extended for
 * customization if desired. Most of the work is handled by <a
 * href="http://hc.apache.org/httpcomponents-client-ga/">Apache HttpClient</a>.
 * <p>
 * There are alternatives to a servlet based proxy such as Apache mod_proxy if
 * that is available to you. However this servlet is easily customizable by
 * Java, secure-able by your web application's security (e.g. spring-security),
 * portable across servlet engines, and is embeddable into another web
 * application.
 * </p>
 * <p>
 * Inspiration: http://httpd.apache.org/docs/2.0/mod/mod_proxy.html
 * </p>
 *
 * @author David Smiley dsmiley@mitre.org
 *         <p/>
 *         Updated for commons-httpclient 4.3 and added multithreaded
 *         authorization, logging
 * @author Connor Barry connor.barry@domain
 */
public class ProxyServlet extends HttpServlet {
    private static final long serialVersionUID = 1375951949525909523L;
    static final Logger log = Logger.getLogger(ProxyServlet.class);

    private static KeyStore KEY_STORE = null;
    private static KeyStore TRUST_STORE = null;
    private static String STORE_SECURE_KEY = "storeSecureKey";

    private static String KEY_STORE_FILE = "keyStoreFile";
    private static String TRUST_STORE_FILE = "trustStoreFile";

    private static String CAPRI_AUTHENTICATION_EXE = "capriAuthenticationExe";

    private static final int BYTES_IN_KILOBYTE = 1024;
    /**
     * We'll stop transmitting data this many bytes before the end of the
     * InputStream if we still haven't been authorized (to avoid any risk the
     * target system will assume a complete response)
     */
    private static final int TRAILING_BYTES_TO_HOLD = 2 * BYTES_IN_KILOBYTE;

    /**
     * The maximum number of seconds we allow the authorization executable to
     * take before we consider it broken and assume authorization failed
     */
    private static final int AUTH_TIMEOUT_SECONDS = 30;

    /**
     * The parameter name for the target (destination) URI to proxy to.
     */
    private static final String SERVLET_PARAM_TARGET_URI = "targetUri";

    /**
     * The HTTP session attribute name for whether the user is authorized
     */
    private static final String HTTP_SESSION_ATTR_AUTHORIZED = "authorized";

    /**
     *
     */
    private static final String HEADER_VISTA_AUTH_TOKEN = "authToken";

    /**
     * These are the "hop-by-hop" headers that should not be copied.
     * http://www.w3.org/Protocols/rfc2616/rfc2616-sec13.html I use an
     * HttpClient HeaderGroup class instead of Set<String> because this approach
     * does case-nsensitive lookup faster.
     */
    protected static final HeaderGroup hopByHopHeaders;

    static {
        hopByHopHeaders = new HeaderGroup();
        String[] headers = new String[]{"Connection", "Keep-Alive",
                "Proxy-Authenticate", "Proxy-Authorization", "TE", "Trailers",
                "Transfer-Encoding", "Upgrade"};

        for (String header : headers)
            hopByHopHeaders.addHeader(new BasicHeader(header, null));
    }

    protected static final BitSet asciiQueryChars;

    static {
        char[] c_unreserved = "_-!.~'()*".toCharArray();
        // plus alphanum

        char[] c_punct = ",;:$&+=".toCharArray();

        char[] c_reserved = "?/[]@".toCharArray();
        // plus punct

        asciiQueryChars = new BitSet(128);
        for (char c = 'a'; c <= 'z'; c++)
            asciiQueryChars.set((int) c);
        for (char c = 'A'; c <= 'Z'; c++)
            asciiQueryChars.set((int) c);
        for (char c = '0'; c <= '9'; c++)
            asciiQueryChars.set((int) c);
        for (char c : c_unreserved)
            asciiQueryChars.set((int) c);
        for (char c : c_punct)
            asciiQueryChars.set((int) c);
        for (char c : c_reserved)
            asciiQueryChars.set((int) c);

        asciiQueryChars.set((int) '%');
        // leave existing percent escapes in place
    }

    /**
     * The URI of the proxy server; this is configurable as a servlet
     * init-param.
     */
    protected URI targetUri;

    /**
     * A thread executor which manages multiple concurrent authorization tasks
     */
    protected ExecutorService executor;

    @Override
    public String getServletInfo() {
        return "A proxy servlet by David Smiley, dsmiley@mitre.org, adapted by Connor Barry, connor.barry@domain";
    }

    @Override
    public void init(ServletConfig servletConfig) throws ServletException {
        super.init(servletConfig);

        try {
            targetUri = new URI(
                    servletConfig.getInitParameter(SERVLET_PARAM_TARGET_URI));
            STORE_SECURE_KEY = servletConfig.getInitParameter(STORE_SECURE_KEY);
            KEY_STORE_FILE = servletConfig.getInitParameter(KEY_STORE_FILE);
            TRUST_STORE_FILE = servletConfig.getInitParameter(TRUST_STORE_FILE);
            CAPRI_AUTHENTICATION_EXE = servletConfig.getInitParameter(CAPRI_AUTHENTICATION_EXE);
        } catch (Exception e) {
            throw new RuntimeException("Error retrieving initial parameter: " + e, e);
        }

        synchronized (this) {
            try {
                KEY_STORE = KeyStore.getInstance(KeyStore.getDefaultType());
                TRUST_STORE = KeyStore.getInstance(KeyStore.getDefaultType());
                log.info("KeyStore and TrustStore instantiated");
            } catch (KeyStoreException e) {
                log.error("KeyStore and TrustStore failed to instantiate", e);
            }

            try {
                KEY_STORE.load(new FileInputStream(KEY_STORE_FILE), STORE_SECURE_KEY.toCharArray());
                TRUST_STORE.load(new FileInputStream(TRUST_STORE_FILE), STORE_SECURE_KEY.toCharArray());
            } catch (IOException e) {
                log.error(e.getMessage(), e);
            } catch (NoSuchAlgorithmException e) {
                log.error(e.getMessage(), e);
            } catch (CertificateException e) {
                log.error(e.getMessage(), e);
            }
        }

		/* Default config; we could use init-params to make this configurable */
        executor = Executors.newCachedThreadPool();
    }

    /**
     * Called from {@link #init(javax.servlet.ServletConfig)}. HttpClient offers
     * many opportunities for customization.
     */
    protected CloseableHttpClient createHttpClient() {
        HttpClientBuilder b = HttpClientBuilder.create();

        /**
         * Setup two-way SSL configuration
         */
        SSLContext sslContext = null;
        try {
            sslContext = SSLContexts.custom().loadTrustMaterial(TRUST_STORE)
                    .loadKeyMaterial(KEY_STORE, STORE_SECURE_KEY.toCharArray()).build();
            log.debug("STORE_SECURE_KEY=" + STORE_SECURE_KEY);
            SSLConnectionSocketFactory sslConnectionSocketFactory = new SSLConnectionSocketFactory(sslContext,
                    SSLConnectionSocketFactory.BROWSER_COMPATIBLE_HOSTNAME_VERIFIER);
            b.setSSLSocketFactory(sslConnectionSocketFactory);
        } catch (Exception e) {
            log.error(e.getMessage(), e);
        }

		/*BasicHttpClientConnectionManager connManager = new BasicHttpClientConnectionManager();
        connManager.setSocketConfig(SocketConfig.custom().setSoKeepAlive(false)
				.setSoLinger(0).setSoReuseAddress(false).build());
		//connManager.setConnectionConfig(connConfig);
		b.setConnectionManager(connManager);//*/

        RequestConfig defaultRequestConfig = RequestConfig.custom()
                .setRedirectsEnabled(false)
                .setStaleConnectionCheckEnabled(true).setConnectTimeout(0)
                .setConnectionRequestTimeout(0).setSocketTimeout(0).build();
        b.setDefaultRequestConfig(defaultRequestConfig);
        /*
         * Automatic retries must be disabled because we only read from the
		 * client request stream once and we don't want to buffer huge multipart
		 * files in memory
		 */
        b.disableAutomaticRetries();
        return b.build();
    }

    @Override
    public void destroy() {
        if (executor != null)
            executor.shutdownNow();
        super.destroy();
    }

    @SuppressWarnings("deprecation")
    @Override
    protected void service(HttpServletRequest servletRequest,
                           HttpServletResponse servletResponse) throws ServletException, IOException {

        log.info("Servlet accessed using " + servletRequest.getMethod() + " method...");

        if (servletRequest.getMethod().equalsIgnoreCase("get")) {
            servletResponse.setContentType("text/html");
            servletResponse.getOutputStream().print("CAPRI Proxy " + Long.toString(System.currentTimeMillis()));
            return;
        }

        if (!servletRequest.getMethod().equalsIgnoreCase("post")) {
            return;
        }

		/*
         * If the user has an existing Tomcat session and they were previously
		 * authenticated, don't force them to re-authenticate
		 */
        final HttpSession session = servletRequest.getSession();
        Boolean authorized = (Boolean) session.getAttribute(HTTP_SESSION_ATTR_AUTHORIZED);
        boolean skipAuth = authorized != null && authorized;
        log.debug("Checking HTTP session for previous authorization: " + Boolean.toString(skipAuth));

        if (true == skipAuth) {
            log.debug("SUCCESS: HTTP session HAS BEEN PREVIOUSLY AUTHORIZED!");
        } else {
            log.debug("SESSIONID: " + servletRequest.getSession().getId());
            log.debug("SESSIONID ISNEW: " + Boolean.toString(servletRequest.getSession().isNew()));
        }

		/*
         * TODO Not ideal design, this is generally designed to be shared by
		 * multiple threads, but I was getting some strange socket-disconnect
		 * errors that need investigating. I'm creating one per-thread as a
		 * workaround for now (overhead should be negligible)
		 */
        CloseableHttpClient proxyClient = createHttpClient();

        HttpRequest proxyRequest = prepareRequest(servletRequest, session, skipAuth);

        copyRequestHeaders(servletRequest, proxyRequest);

        try {
            // Execute the request
            CloseableHttpResponse proxyResponse = proxyClient.execute(
                    URIUtils.extractHost(targetUri), proxyRequest);

            try {
                // Process the response
                int statusCode = proxyResponse.getStatusLine().getStatusCode();

                if (doResponseRedirectOrNotModifiedLogic(servletRequest,
                        servletResponse, proxyResponse, statusCode)) {
                    // just to be sure, but is probably a no-op
                    EntityUtils.consume(proxyResponse.getEntity());
                    return;
                }

				/*
                 * Pass the response code. This method with the "reason phrase"
				 * (e.g. "OK") is deprecated but it's the only way to pass the
				 * reason along too.
				 */
                servletResponse.setStatus(statusCode, proxyResponse
                        .getStatusLine().getReasonPhrase());

                copyResponseHeaders(proxyResponse, servletResponse);

                // Send the content to the client
                String responseEntity = copyResponseEntity(proxyResponse, servletResponse);
                log.info("Servlet response: " + responseEntity);
            } finally {
                proxyResponse.close();
            }
        } catch (Exception e) {
            // abort request, according to best practice with HttpClient
            if (proxyRequest instanceof AbstractExecutionAwareRequest) {
                ((AbstractExecutionAwareRequest) proxyRequest).abort();
            }

            if (e instanceof AuthorizationFailedException) {
                servletResponse.setContentType("text/html");
                if (e.getCause() instanceof TimeoutException) {
                    /*
                     * Auth didn't come back in time; write this to the client
					 * and bail
					 */
                    servletResponse.getOutputStream().print("<proxy-error>AuthTimeout</proxy-error>");
                } else {
                    servletResponse.getOutputStream().print("<proxy-error>ProxyAuthFailed</proxy-error>");
                }
            } else if (e instanceof RuntimeException) {
                throw (RuntimeException) e;
            } else if (e instanceof ServletException) {
                throw (ServletException) e;
            } else if (e instanceof EOFException) {
                servletResponse.getOutputStream().print("<proxy-error>Connection abruptly ended</proxy-error>");
            } else if (e instanceof IOException) {
                throw (IOException) e;
            } else {
                throw new RuntimeException(e);
            }
        } finally {
            try {
                servletResponse.getOutputStream().flush();
            } catch (IOException ignored) {
            }
            proxyClient.close();
        }
    }

    private HttpRequest prepareRequest(HttpServletRequest servletRequest,
                                       final HttpSession session, boolean skipAuth) throws IOException {
        String authToken = servletRequest.getHeader(HEADER_VISTA_AUTH_TOKEN);
        log.debug("credentials: " + authToken);

		/*
		 * Make the Request. note: we won't transfer the protocol version
		 * because I'm not sure it would truly be compatible
		 */
        String method = servletRequest.getMethod();
        String proxyRequestUri = rewriteUrlFromRequest(servletRequest);
        HttpRequest proxyRequest;
		/*
		 * spec: RFC 2616, sec 4.3: either of these two headers signal that
		 * there is a message body.
		 */
        if (servletRequest.getHeader(HttpHeaders.CONTENT_LENGTH) != null
                || servletRequest.getHeader(HttpHeaders.TRANSFER_ENCODING) != null) {
            HttpEntityEnclosingRequest eProxyRequest = new BasicHttpEntityEnclosingRequest(
                    method, proxyRequestUri);
			/*
			 * Add the input entity (streamed). note: we don't bother ensuring
			 * we close the servletInputStream since the container handles it
			 */
            InputStream is = servletRequest.getInputStream();
            if (!skipAuth) {
                ProcessBuilder processBuilder = new ProcessBuilder();
                processBuilder.command(CAPRI_AUTHENTICATION_EXE, authToken);

                /* Kick off activation thread */
                final AuthTask authTask = new AuthTask(processBuilder);
                Future<Boolean> authResult = executor.submit(authTask);

				/*
				 * Wrap the input stream in an AuthorizedInputStream and
				 * register a callback to store the result in the HTTP session
				 */
                @SuppressWarnings("resource")
                AuthorizedInputStream ais = new AuthorizedInputStream(servletRequest.getInputStream(), authResult,
                        AUTH_TIMEOUT_SECONDS, TRAILING_BYTES_TO_HOLD)
                        .addAuthCallback(new AuthCallback() {
                            @Override
                            public void authComplete(boolean isAuthorized) {
                                log.info("isAuthorized request? " + Boolean.toString(isAuthorized));

                                if (isAuthorized) {
                                    session.setAttribute(HTTP_SESSION_ATTR_AUTHORIZED, isAuthorized);
                                    log.debug("Storing authorization result in HTTP session: " + Boolean.toString(isAuthorized));
                                } else {
                                    session.removeAttribute(HTTP_SESSION_ATTR_AUTHORIZED);
                                    log.debug("Removing authorization result in HTTP session: " + Boolean.toString(isAuthorized));
                                }

                                Process process = authTask.getProcess();
                                try {
                                    process.exitValue();
                                } catch (IllegalThreadStateException e) {
                                    log.error("Process has NOT terminated; forcefully killing the process");
                                    process.destroy();
                                }
                            }
                        });
                is = ais;
            } else {
                is = new BufferedInputStream(is);
            }

            InputStreamEntity inputStreamEntity = new InputStreamEntity(is, -1);

            eProxyRequest.setEntity(inputStreamEntity);

            proxyRequest = eProxyRequest;
        } else {
            log.debug("Running basic http request: " + method);
            proxyRequest = new BasicHttpRequest(method, proxyRequestUri);
        }

        return proxyRequest;
    }

    private boolean doResponseRedirectOrNotModifiedLogic(
            HttpServletRequest servletRequest,
            HttpServletResponse servletResponse, HttpResponse proxyResponse,
            int statusCode) throws ServletException, IOException {
		/*
		 * Check if the proxy response is a redirect. The following code is
		 * adapted from org.tigris.noodle.filters.CheckForRedirect
		 */
        if (statusCode >= HttpServletResponse.SC_MULTIPLE_CHOICES /* 300 */
                && statusCode < HttpServletResponse.SC_NOT_MODIFIED /* 304 */) {
            Header locationHeader = proxyResponse
                    .getLastHeader(HttpHeaders.LOCATION);
            if (locationHeader == null) {
                throw new ServletException("Received status code: "
                        + statusCode + " but no " + HttpHeaders.LOCATION
                        + " header was found in the response");
            }
			/*
			 * Modify the redirect to go to this proxy servlet rather that the
			 * proxied host
			 */
            String locStr = rewriteUrlFromResponse(servletRequest,
                    locationHeader.getValue());

            servletResponse.sendRedirect(locStr);
            return true;
        }
		/*
		 * 304 needs special handling. See:
		 * http://www.ics.uci.edu/pub/ietf/http/rfc1945.html#Code304 ; We get a
		 * 304 whenever passed an 'If-Modified-Since' header and the data on
		 * disk has not changed; server responds w/ a 304 saying I'm not going
		 * to send the body because the file has not changed.
		 */
        if (statusCode == HttpServletResponse.SC_NOT_MODIFIED) {
            servletResponse.setIntHeader(HttpHeaders.CONTENT_LENGTH, 0);
            servletResponse.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
            return true;
        }
        return false;
    }

    private void closeQuietly(Closeable closeable) {
        try {
            closeable.close();
        } catch (IOException e) {
            log.debug(e.getMessage(), e);
        }
    }

    /**
     * Copy request headers from the servlet client to the proxy request.
     */
    private void copyRequestHeaders(HttpServletRequest servletRequest,
                                    HttpRequest proxyRequest) {
        // Get an Enumeration of all of the header names sent by the client
        @SuppressWarnings("unchecked")
        Enumeration<String> enumerationOfHeaderNames = servletRequest
                .getHeaderNames();
        while (enumerationOfHeaderNames.hasMoreElements()) {
            String headerName = enumerationOfHeaderNames.nextElement();
            // Instead the content-length is effectively set via
            // InputStreamEntity
            if (headerName.equalsIgnoreCase(HttpHeaders.CONTENT_LENGTH))
                continue;
            if (hopByHopHeaders.containsHeader(headerName))
                continue;

            @SuppressWarnings("unchecked")
            Enumeration<String> headers = servletRequest.getHeaders(headerName);
            while (headers.hasMoreElements()) {// sometimes more than one
                // value
                String headerValue = headers.nextElement();
				/*
				 * In case the proxy host is running multiple virtual servers,
				 * rewrite the Host header to ensure that we get content from
				 * the correct virtual server
				 */
                if (headerName.equalsIgnoreCase(HttpHeaders.HOST)) {
                    HttpHost host = URIUtils.extractHost(targetUri);
                    headerValue = host.getHostName();
                    if (host.getPort() != -1)
                        headerValue += ":" + host.getPort();
                }
                proxyRequest.addHeader(headerName, headerValue);
            }
        }
    }

    /**
     * Copy proxied response headers back to the servlet client.
     */
    private void copyResponseHeaders(HttpResponse proxyResponse,
                                     HttpServletResponse servletResponse) {
        for (Header header : proxyResponse.getAllHeaders()) {
            if (hopByHopHeaders.containsHeader(header.getName()))
                continue;
            servletResponse.addHeader(header.getName(), header.getValue());
        }
    }

    /**
     * Copy response body data (the entity) from the proxy to the servlet
     * client.
     */
    private String copyResponseEntity(HttpResponse proxyResponse,
                                    HttpServletResponse servletResponse) throws IOException {
        StringBuilder stringBuilder = new StringBuilder();
        HttpEntity entity = proxyResponse.getEntity();

        if (entity != null) {
            OutputStream servletOutputStream = servletResponse.getOutputStream();
            InputStream proxyInputStream = proxyResponse.getEntity().getContent();

            try {
                int read = -2;
                while ((read = proxyInputStream.read()) != -1) {
                    stringBuilder.append((char) read);
                    servletOutputStream.write(read);
                }
            } finally {
                closeQuietly(proxyInputStream);
                closeQuietly(servletOutputStream);
            }
        }

        return stringBuilder.toString();
    }

    /**
     * Reads the request URI from {@code servletRequest} and rewrites it,
     * considering {@link #targetUri}. It's used to make the new request.
     */
    private String rewriteUrlFromRequest(HttpServletRequest servletRequest) {
        StringBuilder uri = new StringBuilder(500);
        uri.append(targetUri.toString());
        // Handle the path given to the servlet
        if (servletRequest.getPathInfo() != null) {// ex: /my/path.html
            uri.append(encodeUriQuery(servletRequest.getPathInfo()));
        }
        // Handle the query string
        String queryString = servletRequest.getQueryString();
        // ex:(following '?'): name=value&foo=bar#fragment
        if (queryString != null && queryString.length() > 0) {
            uri.append('?');
            int fragIdx = queryString.indexOf('#');
            String queryNoFrag = (fragIdx < 0 ? queryString : queryString
                    .substring(0, fragIdx));
            uri.append(encodeUriQuery(queryNoFrag));
            if (fragIdx >= 0) {
                uri.append('#');
                uri.append(encodeUriQuery(queryString.substring(fragIdx + 1)));
            }
        }
        return uri.toString();
    }

    /**
     * For a redirect response from the target server, this translates
     * {@code theUrl} to redirect to and translates it to one the original
     * client can use.
     */
    private String rewriteUrlFromResponse(HttpServletRequest servletRequest,
                                          String theUrl) {
        // TODO document example paths
        if (theUrl.startsWith(targetUri.toString())) {
            String curUrl = servletRequest.getRequestURL().toString();
            // no query
            String pathInfo = servletRequest.getPathInfo();
            if (pathInfo != null) {
                assert curUrl.endsWith(pathInfo);
                curUrl = curUrl.substring(0,
                        curUrl.length() - pathInfo.length());
                // take pathInfo off
            }
            theUrl = curUrl + theUrl.substring(targetUri.toString().length());
        }
        return theUrl;
    }

    /**
     * Encodes characters in the query or fragment part of the URI.
     * <p/>
     * <p/>
     * Unfortunately, an incoming URI sometimes has characters disallowed by the
     * spec. HttpClient insists that the outgoing proxied request has a valid
     * URI because it uses Java's {@link URI}. To be more forgiving, we must
     * escape the problematic characters. See the URI class for the spec.
     *
     * @param in example: name=value&foo=bar#fragment
     */
    private static CharSequence encodeUriQuery(CharSequence in) {
        // Note that I can't simply use URI.java to encode because it will
        // escape pre-existing escaped things.
        StringBuilder outBuf = null;
        Formatter formatter = null;
        for (int i = 0; i < in.length(); i++) {
            char c = in.charAt(i);
            boolean escape = true;
            if (c < 128) {
                if (asciiQueryChars.get((int) c)) {
                    escape = false;
                }
            } else if (!Character.isISOControl(c) && !Character.isSpaceChar(c)) {// not-ascii
                escape = false;
            }
            if (!escape) {
                if (outBuf != null)
                    outBuf.append(c);
            } else {
                // escape
                if (outBuf == null) {
                    outBuf = new StringBuilder(in.length() + 5 * 3);
                    outBuf.append(in, 0, i);
                    if (formatter != null)
                        formatter.close();
                    formatter = new Formatter(outBuf);
                }
                // leading %, 0 padded, width 2, capital hex
                formatter.format("%%%02X", (int) c);// TODO
            }
        }
        return outBuf != null ? outBuf : in;
    }

}
