unit ReflectionWrapper;

{===============================================================================
  ReflectionWrapper CodeCR185 -MER
    ReflectionWrapper contains the definition of TReflectionWrapper, a class to
    encapsulate the launching of Reflection.  The goal of this class is to
    allow CAPRI to use Reflection for secure shell sessions

    To allow the instance to be managed, we retain a handle to the launched
    process.  This allows us to set focus on or terminate the instance.  To
    prevent multiple instances from being launched, this class is implemented
    as a singleton.  Regardless of how many times the class is instantiated,
    only one instance (gReflectionWrapperInst) will be created

    WARNING:  Since TReflectionWrapper overrides NewInstance and always returns
    the object gReflectionWrapperInst, NO CLASSES SHOULD BE DERIVED FROM THIS
    CLASS.  Deriaeb classes would always return a type TReflectionWrapper when
    instantiated
===============================================================================}

interface

uses
  Windows, SysUtils, Classes, ShellApi, Forms, Dialogs, IniFiles, Messages;

type
  EExecError = class(Exception);
  TReflectionWrapper = class(TObject)
  private
    FShellExecInfo: TShellExecuteInfo;
    FHost: string;
    FAccessToken: string;
    FUserName: string;
    FAddTildeToToken: boolean;
    FReferenceCount: integer;
    function ExecuteApp(externalApp: string; parameters: string; defaultDir: string): boolean;
    function GetAccessTokenExecString(): string;
    function GetHostConnectString(): string;
    function GetUserNameConnectString(): string;
  public
    property Host: string read FHost write FHost;
    property AccessToken: string read FAccessToken write FAccessToken;
    property UserName: string read FUserName write FUserName;
    property AddTildeToToken: boolean read FAddTildeToToken write FAddTildeToToken;
    class function NewInstance: TObject; override;
    procedure FreeInstance; override;
    function LaunchEmulator(): boolean;
    function TerminateEmulator(): boolean;
    function IsRunning: boolean;
    procedure ReflectionToFront;
  end;

function GetProcessId(hWnd: HWND): DWORD; stdcall; external kernel32 Name 'GetProcessId';
function EnumWindowsCallback(hWnd: HWND; lParam: LPARAM): BOOL; stdcall;

var
  gReflectionWrapperInst: TReflectionWrapper; // a referece for the single
                                              // instance of TReflectionWrapper
  gReflectionProcessID: DWord; // the process ID of the reflection instance.  It
                               // is global for visibility to EnumWindowsCallback

implementation

{===============================================================================
  FreeInstance
    We override FreeInstance to implement a singleton design for this class.
    We maintain a reference counter and only free the instance if the reference
    counter is 0.  If we are freeing the instance, we also need to free the
    handle to the launched process if it has been assigned

    All class cleanup should occur here as opposed to OnDestroy

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
procedure TReflectionWrapper.FreeInstance;
begin
  Dec(FReferenceCount);
  if (FReferenceCount = 0) then
  begin
    if FShellExecInfo.hProcess <> 0 then begin
      CloseHandle(FShellExecInfo.hProcess);
    end;
    gReflectionWrapperInst := nil;
    inherited FreeInstance;
  end;
end;

{===============================================================================
  NewInstance
    We override NewInstance to implement a singleton design for this class.
    If the instance has not been instantiated, we do so.  This function always
    increments the reference counter and returns the instance

    All class initialization should occur here instead of in the constructor

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
class function TReflectionWrapper.NewInstance: TObject;
begin
  if (not Assigned(gReflectionWrapperInst)) then
  begin
    gReflectionWrapperInst := TReflectionWrapper(inherited NewInstance);
    fillChar(gReflectionWrapperInst.FShellExecInfo, SizeOf(gReflectionWrapperInst.FShellExecInfo), 0);
    gReflectionWrapperInst.FHost := '';
    gReflectionWrapperInst.FAccessToken := '';
    gReflectionWrapperInst.FUserName := '';
    gReflectionWrapperInst.FAddTildeToToken := FALSE;
    gReflectionWrapperInst.FReferenceCount := 0;
  end;
  Result := gReflectionWrapperInst;
  Inc(gReflectionWrapperInst.FReferenceCount);
end;

{===============================================================================
  ExecuteApp
    ExecuteApp is a function that will use ShellExec to launch an external
    program

  Parameters
    externalApp:
      A string representing the name of the executable application.  It can
      contain a full path, or alternately, the path can be specified in the
      defaultDir parameter
    parameters:
      A string representing the command line parameters
    defaultDir:
      A string that specifies the working directory.  If this parameter
      is not set, the current directory is used as the working directory

  Return values
    ExecuteAppAndWait returns a Boolean TRUE if the requested application runs
      and a value of FALSE if it fails to execute

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function TReflectionWrapper.ExecuteApp(externalApp: string; parameters: string; defaultDir: string): boolean;
var
  lExitStatus: DWORD;
  lSystemMessage: string[255];
  lErrorMessage: string;
  lLength: integer;
begin
  lExitStatus := 0;
  try
    // Initialize the ShellExecInfo record
    FillChar(FShellExecInfo, SizeOf(FShellExecInfo), 0);
    FShellExecInfo.cbSize := SizeOf(FShellExecInfo);
    // By setting fMask to SEE_MASK_NOCLOSEPROCESS, it will cause a handle to
    // the process to be set in the hProcess member.  We are responsible for
    // closing that process handle
    FShellExecInfo.fMask := SEE_MASK_NOCLOSEPROCESS;
    FShellExecInfo.Wnd := Application.Handle;
    FShellExecInfo.lpFile := PChar(externalApp);
    FShellExecInfo.lpParameters := PChar(parameters);
    FShellExecInfo.lpDirectory := PChar(defaultDir);
    FShellExecInfo.nShow := SW_SHOWNORMAL;
    // Execute the app
    if not(ShellExecuteEx(@FShellExecInfo)) then begin
      lExitStatus := GetLastError();
      raise EExecError.Create('The application ' + externalApp + ' failed to execute');
    end;
    ExecuteApp := TRUE;
  except
    on E:Exception do begin
        ExecuteApp := FALSE;
        lLength := FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, nil, lExitStatus, 0, @lSystemMessage[1], 255, nil);
        SetLength(lSystemMessage, lLength);
        lErrorMessage := 'Exit status: ' + IntToStr(lExitStatus) + ', ' +
          lSystemMessage + #10 + 'Exception: ' + E.Message;
        MessageBox(Application.Handle, PAnsiChar(lErrorMessage), 'Error in ExecuteApp', MB_ICONERROR);
        CloseHandle(FShellExecInfo.hProcess);
    end;
  end;
end;

{===============================================================================
  LaunchEmulator
    LaunchEmulator is a function that will execute Reflection.
    The default terminal emulator is a default install of reflection.  Other
    install locations can be specified in CapriTerminalEmulators.ini

   The default settings for the external app are the equivalent of an INI named
   CapriTerminalEmulators.ini located in the EXE folder with the following
   entries:
    [Config]
    Application=C:\Program Files\Attachmate\Reflection\R2win.exe

  Return values
    LaunchEmulator returns a Boolean TRUE if the requested application
      runs and a value of FALSE if it fails to execute

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function TReflectionWrapper.LaunchEmulator: boolean;
var
  lTargetPath: string;
  lExternalApp: string;
  lIni: TIniFile;
  lParamString: string;
  lExePath: string;
begin
  if IsRunning then begin
    // if it is already running, just bring the existing instance to front
    ReflectionToFront;
    LaunchEmulator := TRUE;
  end else begin
    try
      lExePath := ExtractFilePath(Application.ExeName);
      // get the external app config from the inifile.  If no INI or config
      // information is present, the default values are for a default
      // reflection install
      lIni := TIniFile.Create(lExePath + 'CapriTerminalEmulators.ini');
      lExternalApp := lIni.ReadString('Config', 'Application', 'C:\Program Files\Attachmate\Reflection\R2win.exe');
      lIni.Free;
      // Extract the path.  ShellExecEx apparently does not handle
      // spaces in folder names of filenames with full paths
      lTargetPath := ExtractFilePath(lExternalApp);

      // Build the parameter string.  We first check to see if the config file
      // is present.  If it is, we add command line parameters to use the config
      // file.  This config file will set reflection to terminate on the end of
      // session.  In order to prevent a prompt for a host name, we have to set
      // NOCONNECT when using the config file
      if FileExists(lExePath + 'CAPRISession.r2w') then
        lParamString := '/NOCONNECT /S "' + lExePath + 'CAPRISession.r2w"'
      else
        lParamString := '';
      // Set parameters for secure shell commands
      lParamString := lParamString + ' /VBA ".connectionType = ""SECURE SHELL"""';
      // Set parameters for host connection commands
      lParamString := lParamString + GetHostConnectString;
      // Set parameters for username commands
      lParamString := lParamString + GetUserNameConnectString;
      // Set parameters for SSH config commands
      lParamString := lParamString + ' /VBA ".ConnectionSettings = ""SSHConfigScheme BlankPassword""" /VBA ".Connect"';
      // Set parameters for access token commands
      lParamString := lParamString + GetAccessTokenExecString;

      LaunchEmulator := ExecuteApp(lExternalApp, lParamString, lTargetPath);
    except
      on E:Exception do begin
          LaunchEmulator := FALSE;
          MessageBox(Application.Handle, PAnsiChar(E.Message), 'Error in Reflection Wrapper', MB_ICONERROR);
      end;
    end;
  end;
end;

{===============================================================================
  TerminateEmulator
    TerminateEmulator is a function that will terminate the launched instance of
    Reflection

  Return values
    TerminateEmulator returns a Boolean TRUE unless the call to TerminateProcess
      fails

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function TReflectionWrapper.TerminateEmulator: boolean;
begin
  Result := TRUE;
  if (IsRunning) then begin
    Result := TerminateProcess(FShellExecInfo.hProcess, 10);
  end;
end;

{===============================================================================
  GetAccessTokenExecString
    GetAccessTokenExecString is a function that will create the command line
    parameters to enter an access code in a Reflection session.  Some servers
    strip the leading tilde, so property AddTildeToToken is a flag for when
    an additional leading tilde is required

  Return values
    GetAccessTokenExecString returns a string representing the command line
    parameters to enter an access code in a Reflection session.  If the
    AccessToken is not set, it will return an empty string

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function TReflectionWrapper.GetAccessTokenExecString: string;
var
  lResult: string;
begin
  lResult := '';
  if (FAccessToken <> '') then begin
    lResult := ' /VBA ".WaitForString ""ACCESS CODE:"" " /VBA ".Transmit ""';
    if FAddTildeToToken then
      lResult := lResult + '~';
    lResult := lResult + FAccessToken + '"" & chr(13)"';
  end;
  result := lResult;
end;

{===============================================================================
  GetHostConnectString
    GetHostConnectString is a function that will create the command line
    parameters to set the host in a Reflection session

  Return values
    GetHostConnectString returns a string representing the command line
    parameters to set the host in a Reflection session.  If the
    Host is not set, it will return an empty string and raise an
    exception

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function TReflectionWrapper.GetHostConnectString: string;
var
  lResult: string;
begin
  lResult := '';
  if (FHost <> '') then begin
    lResult := ' /VBA ".ConnectionSettings = ""Host ';
    lResult := lResult + FHost + '"""';
    result := lResult;
  end else begin
    result := lResult;
    raise EExecError.Create('To execute Reflection, a host must be supplied.');
  end;
end;

{===============================================================================
  GetUserNameConnectString
    GetUserNameConnectString is a function that will create the command line
    parameters to set the username in a Reflection session

  Return values
    GetUserNameConnectString returns a string representing the command line
    parameters to set the username in a Reflection session.  If the
    UserName is not set, it will return an empty string and raise an
    exception

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function TReflectionWrapper.GetUserNameConnectString: string;
var
  lResult: string;
begin
  lResult := '';
  if (FUsername <> '') then begin
    lResult := ' /VBA ".ConnectionSettings = ""UserName ';
    lResult := lResult + FUsername + '"""';
    result := lResult;
  end else begin
    result := lResult;
    raise EExecError.Create('To execute Reflection, a username must be supplied.');
  end;
end;

{===============================================================================
  IsRunning
    IsRunning is a function that will determine the current state of the
    launched instance of reflection

  Return values
    IsRunning returns a boolean representing if the launched application is
    currently running

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function TReflectionWrapper.IsRunning: boolean;
var
  lExitStatus: DWord;
begin
  Result := FALSE;
  if FShellExecInfo.hProcess > 0 then begin
    GetExitCodeProcess(FShellExecInfo.hProcess, lExitStatus);
    if (lExitStatus = STILL_ACTIVE) then
      Result := TRUE;
  end;
end;

{===============================================================================
  ReflectionToFront
    ReflectionToFront is a method that uses a windows callback to bring the
    launched instance of Reflection to front.  Since the callback function
    is not a member of class TReflectionWrapper, a global variable representing
    the process ID of the launched Reflection instance is used

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
procedure TReflectionWrapper.ReflectionToFront;
begin
  if (FShellExecInfo.hProcess <> 0) then begin
    gReflectionProcessID := GetProcessId(FShellExecInfo.hProcess);
    // calling EnumWindows, we pass our callback, EnumWindowsCallback which
    // will be called for each top level window
    EnumWindows(@EnumWindowsCallback, 0);
  end;
end;

{===============================================================================
  EnumWindowsCallback
    EnumWindowsCallback is the callback target for our call to EnumWindows in
    ReflectionToFront.  EnumWindowsCallback will be called with a handle to each
    top level window.  We compare the ProcessID to the one we launched, and if
    that matches we then compare the classname to that of the main window of
    Reflection.  If we have a match, we then restore that window and bring it to
    front and terminate the enumeration

  Parameters
    hWnd:
      An HWND representing a handle to a top level window
    lParam:
      An LPARAM representing an application-defined value passed to the callback
      function.  In our case, this is not used

  Return values
    EnumWindowsCallback returns a Boolean TRUE if we want Windows to continue
    the enumeration.  When we want the enumeration to cease, we return FALSE

  Modification history
    04/2011 CodeCR185 -MER Method created
===============================================================================}
function EnumWindowsCallback(hWnd: HWND; lParam: LPARAM): BOOL;
var
  lPID: DWord;
  lClass: array[0..255] of char;
  lString: string;
begin
  Result := TRUE;
  GetWindowThreadProcessId(hWnd, @lPID);
  if (lPID = gReflectionProcessID) then begin
    GetClassName(hWnd, lClass, 255);
    lString := Format('%s', [lClass]);
    if (lString = 'r2Window') then begin
      // restore in case it is minimized
      PostMessage(hWnd, WM_SYSCOMMAND, SC_RESTORE, 0);
      // bring the window to front and activate the window
      SetForegroundWindow(hWnd);
      // stop enumerating windows
      Result := FALSE;
    end;
  end;
end;

end.
